""" file with orca algorithms subclasses """
from typing import List, Literal
from collections.abc import Callable
import numpy as np

from qlauncher.base import Problem, Algorithm, Result, Backend
from qlauncher.exceptions import DependencyError

try:
    from ptseries.algorithms.binary_solvers import BinaryBosonicSolver
except ImportError as e:
    raise DependencyError(e, install_hint='orca', private=True) from e


class OrcaBackend(Backend):
    """ local backend """

    def __init__(self, name: str) -> None:
        super().__init__(name)

    def get_args(self):
        return {}


class BBS(Algorithm):
    """
    Binary Bosonic Solver algorithm class.

    This class represents the Binary Bosonic Solver (BBS) algorithm. BBS is a quantum-inspired algorithm that
    solves optimization problems by mapping them onto a binary bosonic system. It uses a training process
    to find the optimal solution.

    Attributes:
    - learning_rate (float): The learning rate for the algorithm.
    - updates (int): The number of updates to perform during training.
    - tbi_loops (str): The type of TBI loops to use.
    - print_frequency (int): The frequency at which to print updates.
    - logger (Logger): The logger object for logging algorithm information.

    """
    _algorithm_format = 'qubo'

    def __init__(self, format: Literal['qubo', 'qubo_fn'] = 'qubo', **kwargs) -> None:
        super().__init__()
        self._algorithm_format = format
        self.kwargs = kwargs
        self.input_state = self.kwargs.pop('input_state', None)

    def run(self, problem: Problem, backend: OrcaBackend, formatter: Callable[[Problem], np.ndarray]) -> Result:
        # params = {"tbi_type": self.kwarga['tbi_type']}
        # if backend is not None:
        #     params.update(backend.get_args())
        objective = formatter(problem)
        # TODO: use offset somehow
        if not callable(objective):
            objective, offset = objective
            if self.input_state is None:
                self.input_state = [not i % 2 for i in range(len(objective))]

        self.bbs = BinaryBosonicSolver(
            len(self.input_state),
            objective,
            self.input_state,
            **self.kwargs
        )
        self.bbs.train(**self.kwargs)

        return self.construct_results(self.bbs)

    def get_bitstring(self, result: List[float]) -> str:
        return ''.join(map(str, map(int, result)))

    def construct_results(self, results: BinaryBosonicSolver) -> Result:
        # TODO: add support for distribution (probably with different logger)
        best_bitstring = ''.join(
            map(str, map(int, results.config_min_encountered)))
        best_energy = results.E_min_encountered
        most_common_bitstring = None
        most_common_bitstring_energy = None
        distribution = None
        energy = None
        num_of_samples = results.n_samples
        average_energy = None
        energy_std = None
        #! Todo: instead of None attach relevant info from 'results'
        # results fail to pickle correctly btw
        return Result(best_bitstring, best_energy, most_common_bitstring,
                      most_common_bitstring_energy, distribution, energy,
                      num_of_samples, average_energy, energy_std, None)



""" Algorithms for Qiskit routines """
import json
from datetime import datetime
from collections.abc import Callable

import numpy as np

from qiskit import qpy, QuantumCircuit
from qiskit.circuit.library import PauliEvolutionGate

from qiskit.primitives import PrimitiveResult, SamplerPubResult, BaseEstimatorV1, BaseSamplerV1
from qiskit.primitives.containers import BitArray
from qiskit.primitives.base.base_primitive import BasePrimitive

from qiskit.quantum_info import SparsePauliOp

from qiskit_algorithms.minimum_eigensolvers import QAOA as QiskitQAOA
from qiskit_algorithms.minimum_eigensolvers import SamplingVQEResult
from qiskit_algorithms.optimizers import Optimizer, COBYLA

from qlauncher.base import Problem, Algorithm, Result
from qlauncher.base.base import Backend
from qlauncher.routines.cirq_routines import CirqBackend
from qlauncher.routines.qiskit_routines.backends.qiskit_backend import QiskitBackend


class QiskitOptimizationAlgorithm(Algorithm):
    """ Abstract class for Qiskit optimization algorithms """

    def make_tag(self, problem: Problem, backend: QiskitBackend) -> str:
        tag = problem.__class__.__name__ + '-' + \
            backend.__class__.__name__ + '-' + \
            self.__class__.__name__ + '-' + \
            datetime.today().strftime('%Y-%m-%d')
        return tag

    def get_processing_times(self, tag: str, primitive: BasePrimitive) -> None | tuple[list, list, int]:
        timestamps = []
        usages = []
        qpu_time = 0
        if hasattr(primitive, 'session'):
            jobs = primitive.session.service.jobs(limit=None, job_tags=[tag])
            for job in jobs:
                m = job.metrics()
                timestamps.append(m['timestamps'])
                usages.append(m['usage'])
                qpu_time += m['usage']['quantum_seconds']
        return timestamps, usages, qpu_time


def commutator(op_a: SparsePauliOp, op_b: SparsePauliOp) -> SparsePauliOp:
    """ Commutator """
    return op_a @ op_b - op_b @ op_a


class QAOA(QiskitOptimizationAlgorithm):
    """Algorithm class with QAOA.

    Args:
        p (int): The number of QAOA steps. Defaults to 1.
        optimizer (Optimizer | None): Optimizer used during algorithm runtime. If set to `None` turns into COBYLA. Defaults to None,
        alternating_ansatz (bool): Whether to use an alternating ansatz. Defaults to False. If True, it's recommended to provide a mixer_h to alg_kwargs.
        aux: Auxiliary input for the QAOA algorithm.
        **alg_kwargs: Additional keyword arguments for the base class.

    Attributes:
        name (str): The name of the algorithm.
        aux: Auxiliary input for the QAOA algorithm.
        p (int): The number of QAOA steps.
        optimizer (Optimizer): Optimizer used during algorithm runtime.
        alternating_ansatz (bool): Whether to use an alternating ansatz.
        parameters (list): List of parameters for the algorithm.
        mixer_h (SparsePauliOp | None): The mixer Hamiltonian.

    """
    _algorithm_format = 'hamiltonian'

    def __init__(self, p: int = 1, optimizer: Optimizer | None = None, alternating_ansatz: bool = False, aux=None, **alg_kwargs):
        super().__init__(**alg_kwargs)
        self.name: str = 'qaoa'
        self.aux = aux
        self.p: int = p
        self.optimizer: Optimizer = optimizer if optimizer is not None else COBYLA()
        self.alternating_ansatz: bool = alternating_ansatz
        self.parameters = ['p']
        self.mixer_h: SparsePauliOp | None = None
        self.initial_state: QuantumCircuit | None = None

    @property
    def setup(self) -> dict:
        return {
            'aux': self.aux,
            'p': self.p,
            'parameters': self.parameters,
            'arg_kwargs': self.alg_kwargs
        }

    def parse_samplingVQEResult(self, res: SamplingVQEResult, res_path) -> dict:
        res_dict = {}
        for k, v in vars(res).items():
            if k[0] == "_":
                key = k[1:]
            else:
                key = k
            try:
                res_dict = {**res_dict, **json.loads(json.dumps({key: v}))}
            except TypeError as ex:
                if str(ex) == 'Object of type complex128 is not JSON serializable':
                    res_dict = {**res_dict, **
                                json.loads(json.dumps({key: v}, default=repr))}
                elif str(ex) == 'Object of type ndarray is not JSON serializable':
                    res_dict = {**res_dict, **
                                json.loads(json.dumps({key: v}, default=repr))}
                elif str(ex) == 'keys must be str, int, float, bool or None, not ParameterVectorElement':
                    res_dict = {**res_dict, **
                                json.loads(json.dumps({key: repr(v)}))}
                elif str(ex) == 'Object of type OptimizerResult is not JSON serializable':
                    # recursion ftw
                    new_v = self.parse_samplingVQEResult(v, res_path)
                    res_dict = {**res_dict, **
                                json.loads(json.dumps({key: new_v}))}
                elif str(ex) == 'Object of type QuantumCircuit is not JSON serializable':
                    path = res_path + '.qpy'
                    with open(path, 'wb') as f:
                        qpy.dump(v, f)
                    res_dict = {**res_dict, **{key: path}}
        return res_dict

    def run(self, problem: Problem, backend: Backend, formatter: Callable) -> Result:
        """ Runs the QAOA algorithm """
        if not (isinstance(backend, QiskitBackend) or isinstance(backend, CirqBackend)):
            raise ValueError('Backend should be CirqBackend, QiskitBackend or subclass.')
        hamiltonian: SparsePauliOp = formatter(problem)
        energies = []

        def qaoa_callback(evaluation_count, params, mean, std):
            energies.append(mean)

        tag = self.make_tag(problem, backend)
        sampler = backend.samplerV1
        # sampler.set_options(job_tags=[tag])

        if self.alternating_ansatz:
            if self.mixer_h is None:
                self.mixer_h = formatter.get_mixer_hamiltonian(problem)
            if self.initial_state is None:
                self.initial_state = formatter.get_QAOAAnsatz_initial_state(
                    problem)

        qaoa = QiskitQAOA(sampler, self.optimizer, reps=self.p, callback=qaoa_callback,
                          mixer=self.mixer_h, initial_state=self.initial_state, **self.alg_kwargs)
        qaoa_result = qaoa.compute_minimum_eigenvalue(hamiltonian, self.aux)
        depth = qaoa.ansatz.decompose(reps=10).depth()
        if 'cx' in qaoa.ansatz.decompose(reps=10).count_ops():
            cx_count = qaoa.ansatz.decompose(reps=10).count_ops()['cx']
        else:
            cx_count = 0
        timestamps, usages, qpu_time = self.get_processing_times(tag, sampler)
        return self.construct_result({'energy': qaoa_result.eigenvalue,
                                      'depth': depth,
                                      'cx_count': cx_count,
                                      'qpu_time': qpu_time,
                                      'energies': energies,
                                      'SamplingVQEResult': qaoa_result,
                                      'usages': usages,
                                      'timestamps': timestamps})

    def construct_result(self, result: dict) -> Result:

        best_bitstring = self.get_bitstring(result)
        best_energy = result['energy']

        distribution = dict(result['SamplingVQEResult'].eigenstate.items())
        most_common_value = max(
            distribution, key=distribution.get)
        most_common_bitstring = bin(most_common_value)[2:].zfill(
            len(best_bitstring))
        most_common_bitstring_energy = distribution[most_common_value]
        num_of_samples = 0  # TODO: implement
        average_energy = np.mean(result['energies'])
        energy_std = np.std(result['energies'])
        return Result(best_bitstring, best_energy, most_common_bitstring, most_common_bitstring_energy, distribution, result['energies'], num_of_samples, average_energy, energy_std, result)

    def get_bitstring(self, result) -> str:
        return result['SamplingVQEResult'].best_measurement['bitstring']


class FALQON(QiskitOptimizationAlgorithm):
    """ 
    Algorithm class with FALQON.

    Args:
        driver_h (Optional[Operator]): The driver Hamiltonian for the problem.
        delta_t (float): The time step for the evolution operators.
        beta_0 (float): The initial value of beta.
        n (int): The number of iterations to run the algorithm.
        **alg_kwargs: Additional keyword arguments for the base class.

    Attributes:
        driver_h (Optional[Operator]): The driver Hamiltonian for the problem.
        delta_t (float): The time step for the evolution operators.
        beta_0 (float): The initial value of beta.
        n (int): The number of iterations to run the algorithm.
        cost_h (Optional[Operator]): The cost Hamiltonian for the problem.
        n_qubits (int): The number of qubits in the problem.
        parameters (List[str]): The list of algorithm parameters.

    """
    _algorithm_format = 'hamiltonian'

    def __init__(
        self,
        driver_h: SparsePauliOp | None = None,
        delta_t: float = 0.03,
        beta_0: float = 0.0,
        max_reps: int = 20
    ) -> None:
        super().__init__()
        self.driver_h = driver_h
        self.cost_h = None
        self.delta_t = delta_t
        self.beta_0 = beta_0
        self.max_reps = max_reps
        self.n_qubits: int = 0
        self.parameters = ['n', 'delta_t', 'beta_0']

    @property
    def setup(self) -> dict:
        return {
            'driver_h': self.driver_h,
            'delta_t': self.delta_t,
            'beta_0': self.beta_0,
            'n': self.max_reps,
            'cost_h': self.cost_h,
            'n_qubits': self.n_qubits,
            'parameters': self.parameters,
            'arg_kwargs': self.alg_kwargs
        }

    def _get_path(self) -> str:
        return f'{self.name}@{self.max_reps}@{self.delta_t}@{self.beta_0}'

    def run(self, problem: Problem, backend: QiskitBackend, formatter: Callable) -> Result:
        """ Runs the FALQON algorithm """

        if isinstance(backend.sampler, BaseSamplerV1) or isinstance(backend.estimator, BaseEstimatorV1):
            raise ValueError("FALQON works only on V2 samplers and estimators, consider using a different backend.")

        cost_h = formatter(problem)

        if cost_h is None:
            raise ValueError("Formatter returned None")

        self.n_qubits = cost_h.num_qubits

        best_sample, betas, energies, depths, cnot_counts = self._falqon_subroutine(cost_h, backend)

        best_data: BitArray = best_sample[0].data.meas
        counts: dict = best_data.get_counts()
        shots = best_data.num_shots

        result = {'betas': betas,
                  'energies': energies,
                  'depths': depths,
                  'cxs': cnot_counts,
                  'n': self.max_reps,
                  'delta_t': self.delta_t,
                  'beta_0': self.beta_0,
                  'energy': min(energies),
                  }

        return Result(
            best_bitstring=max(counts, key=counts.get),
            most_common_bitstring=max(counts, key=counts.get),
            distribution={k: v/shots for k, v in counts.items()},
            energies=energies,
            energy_std=np.std(energies),
            best_energy=min(energies),
            num_of_samples=shots,
            average_energy=np.mean(energies),
            most_common_bitstring_energy=0,
            result=result
        )

    def _add_ansatz_part(
        self,
        cost_hamiltonian: SparsePauliOp,
        driver_hamiltonian: SparsePauliOp,
        beta: float,
        circuit: QuantumCircuit
    ) -> None:
        """Adds a single FALQON ansatz 'building block' with the specified beta to the circuit"""
        circ_part = QuantumCircuit(circuit.num_qubits)

        circ_part.append(PauliEvolutionGate(cost_hamiltonian, time=self.delta_t), circ_part.qubits)
        circ_part.append(PauliEvolutionGate(driver_hamiltonian, time=self.delta_t * beta), circ_part.qubits)

        circuit.compose(circ_part, circ_part.qubits, inplace=True)

    def _build_ansatz(self, cost_hamiltonian, driver_hamiltonian, betas):
        """Build the FALQON circuit for the given betas"""

        circ = QuantumCircuit(self.n_qubits)
        circ.h(range(self.n_qubits))

        for beta in betas:
            circ.append(PauliEvolutionGate(cost_hamiltonian, time=self.delta_t), circ.qubits)
            circ.append(PauliEvolutionGate(driver_hamiltonian, time=self.delta_t * beta), circ.qubits)
        return circ

    def _falqon_subroutine(
            self,
            cost_hamiltonian: SparsePauliOp,
            backend: QiskitBackend
    ) -> tuple[PrimitiveResult[SamplerPubResult], list[float], list[float], list[int], list[int]]:
        """
        Run the 'meat' of the algorithm.

        Args:
            cost_hamiltonian (SparsePauliOp): Cost hamiltonian from the formatter.
            backend (QiskitBackend): Backend

        Returns:
            tuple[PrimitiveResult[SamplerPubResult], list[float], list[float], list[int], list[int]]: 
            Sampler result from best betas, list of betas, list of energies, list of depths, list of cnot counts
        """

        if self.driver_h is None:
            self.driver_h = SparsePauliOp.from_sparse_list([("X", [i], 1) for i in range(self.n_qubits)], num_qubits=self.n_qubits)
            driver_hamiltonian = self.driver_h
        else:
            driver_hamiltonian = self.driver_h

        hamiltonian_commutator = complex(0, 1) * commutator(driver_hamiltonian, cost_hamiltonian)

        betas = [self.beta_0]
        energies = []
        cnot_counts = []
        circuit_depths = []

        circuit = QuantumCircuit(self.n_qubits)
        circuit.h(circuit.qubits)

        self._add_ansatz_part(cost_hamiltonian, driver_hamiltonian, self.beta_0, circuit)

        for i in range(self.max_reps):

            beta = -1 * backend.estimator.run([(circuit, hamiltonian_commutator)]).result()[0].data.evs
            betas.append(beta)

            self._add_ansatz_part(cost_hamiltonian, driver_hamiltonian, beta, circuit)

            energy = backend.estimator.run([(circuit, cost_hamiltonian)]).result()[0].data.evs
            # print(i, energy)
            energies.append(energy)
            circuit_depths.append(circuit.depth())
            cnot_counts.append(circuit.count_ops().get('cx', 0))

        argmin = np.argmin(np.asarray(energies))

        sampling_circuit = self._build_ansatz(cost_hamiltonian, driver_hamiltonian, betas[:argmin])
        sampling_circuit.measure_all()

        best_sample = backend.sampler.run([(sampling_circuit)]).result()

        return best_sample, betas, energies, circuit_depths, cnot_counts



"""  Module for Job Shop Scheduling Problem (JSSP)."""
from collections import defaultdict
from typing import Dict, List, Literal, Tuple
try:
    from qlauncher.problems.problem_formulations.jssp.qiskit_scheduler import get_jss_hamiltonian
except ModuleNotFoundError:
    pass
from qlauncher.base import Problem


class JSSP(Problem):
    """
    Class for Job Shop Scheduling Problem.

    This class represents Job Shop Scheduling Problem (JSSP) which is a combinatorial optimization problem that involves 
    scheduling a set of jobs on a set of machines. Each job consists of a sequence of operations that must be performed 
    on different machines. The objective is to find a schedule that minimizes the makespan, i.e., the total time required
    to complete all jobs. The class contains an instance of the problem, so it can be passed into Quantum Launcher.


    Attributes:
        max_time (int): The maximum time for the scheduling problem.
        onehot (str): The one-hot encoding method to be used.
        optimization_problem (bool): Flag indicating whether the problem is an optimization problem or a decision problem.
        results (dict): Dictionary to store the results of the problem instance.

    """

    def __init__(self, max_time: int, instance: Dict[str, List[Tuple[str, int]]],
                 instance_name: str = 'unnamed', optimization_problem: bool = False, onehot: Literal['exact', 'quadratic'] = 'exact') -> None:
        super().__init__(instance=instance, instance_name=instance_name)
        self.max_time = max_time
        self.onehot = onehot
        self.optimization_problem = optimization_problem

        self.h_d, self.h_o, self.h_pos_by_label, self.h_label_by_pos = get_jss_hamiltonian(self.instance, max_time,
                                                                                           onehot)

        self.results = {'instance_name': instance_name,
                        'max_time': max_time,
                        'onehot': onehot,
                        'H_pos_by_label': self.h_pos_by_label,
                        'H_label_by_pos': self.h_label_by_pos}

        self.variant = 'optimization' if optimization_problem else 'decision'

    @property
    def setup(self) -> dict:
        return {
            'max_time': self.max_time,
            'onehot': self.onehot,
            'optimization_problem': self.optimization_problem,
            'instance_name': self.instance_name
        }

    def _get_path(self) -> str:
        return f'{self.name}@{self.instance_name}@{self.max_time}@{"optimization" if self.optimization_problem else "decision"}@{self.onehot}'

    @staticmethod
    def from_preset(instance_name: str, **kwargs) -> "JSSP":
        match instance_name:
            case 'default':
                max_time = 3
                instance = {"cupcakes": [("mixer", 2), ("oven", 1)],
                            "smoothie": [("mixer", 1)],
                            "lasagna": [("oven", 2)]}
            case _:
                raise ValueError(f"Instance {instance_name} does not exist choose instance_name from the following: ('toy')")

        return JSSP(max_time=max_time, instance=instance, instance_name=instance_name, **kwargs)

    @classmethod
    def from_file(cls, path: str, **kwargs) -> "JSSP":
        job_dict = defaultdict(list)
        with open(path, 'r', encoding='utf-8') as file_:
            file_.readline()
            for i, line in enumerate(file_):
                lint = list(map(int, line.split()))
                job_dict[i + 1] = [x for x in
                                   zip(lint[::2],  # machines
                                       lint[1::2]  # operation lengths
                                       )]
        return JSSP(instance=job_dict, **kwargs)



"""
Routine file for Cirq library
"""
from typing import Literal
from collections.abc import Sequence

import qiskit
from qiskit.primitives import Sampler
from qiskit.primitives.base.sampler_result import SamplerResult
from qiskit.result import QuasiDistribution

from qlauncher.base import Backend
from qlauncher.base.translator import Translation
from qlauncher.exceptions import DependencyError
try:
    import cirq
    from cirq.sim.sparse_simulator import Simulator
except ImportError as e:
    raise DependencyError(e, install_hint='cirq') from e


class _CirqRunner:
    simulator = Simulator()
    repetitions = 1024

    @classmethod
    def calculate_circuit(cls, circuit: qiskit.QuantumCircuit) -> dict:
        circuit = circuit.measure_all(inplace=False)
        cirq_circ = Translation.get_translation(circuit, 'cirq')
        result = cls.simulator.run(cirq_circ, repetitions=cls.repetitions)
        return cls._result_to_dist(result)

    @classmethod
    def _result_to_dist(cls, result) -> dict:
        return cirq_result_to_probabilities(result)


class CirqSampler(Sampler):
    """ Sampler adapter for Cirq """

    def _call(self, circuits: Sequence[int], parameter_values: Sequence[Sequence[float]], **run_options) -> SamplerResult:
        bound_circuits = []
        for i, value in zip(circuits, parameter_values):
            bound_circuits.append(
                self._circuits[i]
                if len(value) == 0
                else self._circuits[i].assign_parameters(dict(zip(self._parameters[i], value)))
            )
        distributions = [_CirqRunner.calculate_circuit(circuit) for circuit in bound_circuits]
        quasi_dists = list(map(QuasiDistribution, distributions))
        return SamplerResult(quasi_dists, [{} for _ in range(len(parameter_values))])


class CirqBackend(Backend):
    """

    Args:
        Backend (_type_): _description_
    """

    def __init__(self, name: Literal['local_simulator'] = 'local_simulator'):
        self.sampler = self.samplerV1 = CirqSampler()
        super().__init__(name)


def cirq_result_to_probabilities(
    result: cirq.Result,
    integer_keys: bool = False
) -> dict:
    measurements = result.measurements

    sorted_keys = list(measurements.keys())

    bitstrings = []
    num_shots = len(measurements[sorted_keys[0]])

    for shot_index in range(num_shots):
        bits = []
        for key in sorted_keys:
            bits.extend(str(b) for b in measurements[key][shot_index])

        bitstring = "".join(bits)
        bitstrings.append(bitstring)

    counts = {}
    for bs in bitstrings:
        counts[bs] = counts.get(bs, 0) + 1

    total_shots = sum(counts.values())
    if integer_keys:
        prob_dict = {int(k, 2): v / total_shots for k, v in counts.items()}
    else:
        prob_dict = {k: v / total_shots for k, v in counts.items()}

    return prob_dict



from typing import Sequence

from qiskit import transpile
from qiskit.result import QuasiDistribution
from qiskit.primitives import SamplerResult, BasePrimitiveJob, BitArray
from qiskit.primitives.base import BaseSamplerV1, BaseSamplerV2


class RuntimeJobV2Adapter(BasePrimitiveJob):
    def __init__(self, job, **kwargs):
        super().__init__(job.job_id(), **kwargs)
        self.job = job

    def result(self):
        raise NotImplementedError()

    def cancel(self):
        return self.job.cancel()

    def status(self):
        return self.job.status()

    def done(self):
        return self.job.done()

    def cancelled(self):
        return self.job.cancelled()

    def running(self):
        return self.job.running()

    def in_final_state(self):
        return self.job.in_final_state()


class SamplerV2JobAdapter(RuntimeJobV2Adapter):
    """
    Dummy data holder, returns a v1 SamplerResult from v2 sampler job.
    """

    def __init__(self, job, **kwargs):
        super().__init__(job, **kwargs)

    def _get_quasi_meta(self, res):
        data = BitArray.concatenate_bits(list(res.data.values()))
        counts = data.get_int_counts()
        probs = {k: v/data.num_shots for k, v in counts.items()}
        quasi_dists = QuasiDistribution(probs, shots=data.num_shots)

        metadata = res.metadata
        metadata["sampler_version"] = 2  # might be useful for debugging

        return quasi_dists, metadata

    def result(self):
        res = self.job.result()
        qd, metas = [], []
        for r in res:
            quasi_dist, metadata = self._get_quasi_meta(r)
            qd.append(quasi_dist)
            metas.append(metadata)

        return SamplerResult(quasi_dists=qd, metadata=metas)


def _transpile_circuits(circuits, backend):
    # Transpile qaoa circuit to backend instruction set, if backend is provided
    # ? I pass a backend into SamplerV2 as *mode* but here sampler_v2.mode returns None, why?
    if not backend is None:
        if isinstance(circuits, Sequence):
            circuits = [transpile(circuit) for circuit in circuits]
        else:
            circuits = transpile(circuits)

    return circuits


class SamplerV2Adapter(BaseSamplerV1):
    """
    V1 adapter for V2 samplers.
    """

    def __init__(self, sampler_v2: BaseSamplerV2, backend=None):
        """
        Args:
            sampler_v2 (BaseSamplerV2): V2 sampler to be adapted.
            backend (Optional[Backend]): Backend to transpile circuits to.
        """
        self.sampler_v2 = sampler_v2
        self.backend = backend
        super().__init__()

    def _run(self, circuits, parameter_values=None, **run_options) -> SamplerV2JobAdapter:
        circuits = _transpile_circuits(circuits, self.backend)
        v2_list = list(zip(circuits, parameter_values))
        job = self.sampler_v2.run(pubs=v2_list, **run_options)

        return SamplerV2JobAdapter(job)



""" File with templates """
from typing import Any, Literal, Optional, Union
import json
import pickle
import logging
from qlauncher.base.adapter_structure import get_formatter, ProblemFormatter
from qlauncher.base import Problem, Algorithm, Backend, Result
from qlauncher.problems import Raw


class QuantumLauncher:
    """
    Quantum Launcher class.

    Quantum launcher is used to run quantum algorithms on specific problem instances and backends.
    It provides methods for binding parameters, preparing the problem, running the algorithm, and processing the results.

    Attributes:
        problem (Problem): The problem instance to be solved.
        algorithm (Algorithm): The quantum algorithm to be executed.
        backend (Backend, optional): The backend to be used for execution. Defaults to None.
        path (str): The path to save the results. Defaults to 'results/'.
        binding_params (dict or None): The parameters to be bound to the problem and algorithm. Defaults to None.
        encoding_type (type): The encoding type to be used changing the class of the problem. Defaults to None.

    Example of usage::

            from templates import QuantumLauncher
            from problems import MaxCut
            from qiskit_routines import QAOA, QiskitBackend

            problem = MaxCut(instance_name='default')
            algorithm = QAOA()
            backend = QiskitBackend('local_simulator')

            launcher = QuantumLauncher(problem, algorithm, backend)
            result = launcher.process(save_pickle=True)
            print(result)

    """

    def __init__(self, problem: Problem, algorithm: Algorithm, backend: Backend = None, logger: Optional[logging.Logger] = None) -> None:

        if not isinstance(problem, Problem):
            problem = Raw(problem)

        self.problem: Problem = problem
        self.algorithm: Algorithm = algorithm
        self.backend: Backend = backend
        self.formatter: ProblemFormatter = get_formatter(self.problem._problem_id, self.algorithm._algorithm_format)

        if logger is None:
            logger: logging.Logger = logging.getLogger('QuantumLauncher')
        self.logger = logger

        # logging.info(f'Found proper formatter, with formatter structure: {self.formatter.get_pipeline()}')

        self.res: dict = {}

    def run(self, **kwargs) -> Result:
        """
        Finds proper formatter, and runs the algorithm on the problem with given backends.

        Returns:
            dict: The results of the algorithm execution.
        """

        self.formatter.set_run_params(kwargs)

        self.result = self.algorithm.run(self.problem, self.backend, formatter=self.formatter)
        logging.info('Algorithm ended successfully!')
        return self.result

    def save(self, path: str, format: Literal['pickle', 'txt', 'json'] = 'pickle'):
        logging.info(f'Saving results to file: {path}')
        if format == 'pickle':
            with open(path, mode='wb') as f:
                pickle.dump(self.result, f)
        elif format == 'json':
            with open(path, mode='w', encoding='utf-8') as f:
                json.dump(self.result.__dict__, f, default=fix_json)
        elif format == 'txt':
            with open(path, mode='w', encoding='utf-8') as f:
                f.write(self.result.__str__())
        else:
            raise ValueError(
                f'format: {format} in not supported try: pickle, txt, csv or json')

    def process(self, *, file_path: Optional[str] = None, format: Union[Literal['pickle', 'txt', 'json'], list[Literal['pickle', 'txt', 'json']]] = 'pickle', **kwargs) -> dict:
        """
        Runs the algorithm, processes the data, and saves the results if specified.

        Args:
            file_path (Optional[str]): Flag indicating whether to save the results to a file. Defaults to None.
            format (Union[Literal['pickle', 'txt', 'json'], list[Literal['pickle', 'txt', 'json']]]): Format in which file should be saved. Defaults to 'pickle'

        Returns:
            dict: The processed results.
        """
        results = self.run(**kwargs)
        energy = results.result['energy']
        res = {}
        res['problem_setup'] = self.problem.setup
        res['algorithm_setup'] = self.algorithm.setup
        res['algorithm_setup']['variant'] = self.problem.variant
        res['backend_setup'] = self.backend.setup
        res['results'] = results

        self._file_name = self.problem.path + '-' + \
            self.backend.path + '-' \
            + self.algorithm.path + '-' + str(energy)

        if file_path is not None and isinstance(format, str):
            self.save(file_path, format)
        if file_path is not None and isinstance(format, list):
            for form in format:
                self.save(file_path, form)
        return res


def fix_json(o: object):
    # if o.__class__.__name__ == 'SamplingVQEResult':
    #     parsed = self.algorithm.parse_samplingVQEResult(o, self._full_path)
    #     return parsed
    if o.__class__.__name__ == 'complex128':
        return repr(o)
    print(
        f'Name of object {o.__class__} not known, returning None as a json encodable')
    return None
